/*
 *  Copyright (C) 2002 Tomasz Kojm <zolw@konarski.edu.pl>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Sat May 18 15:23:21 CEST 2002: included cpu autodetection from Magnus Ekdahl
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/types.h>
#include <signal.h>
#include <clamav.h>

#include "defaults.h"
#include "others.h"
#include "options.h"
#include "manager.h"
#include "treewalk.h"
#include "shared.h"
#include "mbox.h"

#ifdef C_LINUX
dev_t procdev;
#endif

int scanmanager(const struct optstruct *opt)
{
	mode_t fmode;
	int ret = 0, threads = 0, compression = 0, fmodeint;
	struct cl_node *trie = NULL;
	struct cl_limits *limits = NULL;
	struct passwd *user = NULL;
	struct stat sb;
	char *fullpath = NULL, cwd[200];
#ifdef CLAMSCAN_THREADS
	struct sigaction act;
#endif

/* njh@bandsman.co.uk: BeOS */
#if !defined(C_CYGWIN) && !defined(C_BEOS)
    if(!getuid()) {
	if((user = getpwnam(UNPUSER)) == NULL) {
	    mprintf("@Can't get information about user "UNPUSER".\n");
	    exit(60); /* this is critical problem, so we just exit here */
	}
    }
#endif

    if(optl(opt, "threads")) {
	if(!isnumb(getargl(opt, "threads")) || (threads = atoi(getargl(opt, "threads"))) < 0) {
	    mprintf("@Invalid threads number.\n");
	    return 51;
	}
    }
#ifdef C_LINUX
    else /* Autodetecting threads */
      {
	if((threads = detectcpu()) > 1)
	    mprintf("Autodetected %d CPUs. Starting %d threads.\n", threads, threads);
	else
	    threads = 0;
      }
#endif

    if(optl(opt, "unzip") || optl(opt, "unrar") || optl(opt, "unace") ||
       optl(opt, "unarj") || optl(opt, "zoo") || optl(opt, "jar") ||
       optl(opt, "lha") || optl(opt, "tar") || optl(opt, "tgz") ||
       optl(opt, "deb"))
	    compression = 1;


    /* now initialize the database */

    if(optc(opt, 'd')) {
	stat(getargc(opt, 'd'), &sb);
	switch(sb.st_mode & S_IFMT) {
	    case S_IFREG:
		if((ret = cl_loaddb(getargc(opt, 'd'), &trie, &claminfo.signs))) {
		    mprintf("@%s\n", cl_perror(ret));
		    return 50;
		}
		break;
            case S_IFDIR:
		if((ret = cl_loaddbdir(getargc(opt, 'd'), &trie, &claminfo.signs))) {
		    mprintf("@%s\n", cl_perror(ret));
		    return 50;
		}
		break;
            default:
		mprintf("@%s: Not supported database file type\n", getargc(opt, 'd'));
		return 50;
	}

    } else {
	if((ret = cl_loaddbdir(cl_retdbdir(), &trie, &claminfo.signs))) {
	    mprintf("@%s\n", cl_perror(ret));
	    return 50;
	}
    }


    if(!trie) {
	mprintf("@Can't initialize the virus database.\n");
	return 50;
    }

    /* build the proper trie */
    cl_buildtrie(trie);

#ifdef CLAMSCAN_THREADS
    act.sa_handler = thr_exit;
    sigfillset(&(act.sa_mask));
    sigaction(SIGUSR1, &act, NULL);
    thr_pid = getpid();
#endif

    /* parse limits */


    if(optl(opt, "max-files") || optl(opt, "max-space") || optl(opt, "max-recursion")) {
    	limits = (struct cl_limits *) calloc(1, sizeof(struct cl_limits));

	if(optl(opt, "max-space")) {
		char *cpy, *ptr;
	    ptr = getargl(opt, "max-space");
	    if(tolower(ptr[strlen(ptr) - 1]) == 'm') {
	        cpy = mcalloc(strlen(ptr), sizeof(char));
	        strncpy(cpy, ptr, strlen(ptr) - 1);
	        limits->maxfilesize = atoi(cpy) * 1024 * 1024;
	        free(cpy);
	    } else
	        limits->maxfilesize = atoi(ptr) * 1024;
	} else
	    limits->maxfilesize = 0;

	if(optl(opt, "max-files"))
	    limits->maxfiles = atoi(getargl(opt, "max-files"));

	if(optl(opt, "max-recursion"))
	    limits->maxreclevel = atoi(getargl(opt, "max-recursion"));

    }

#ifdef C_LINUX
    if(stat("/proc", &sb) == -1)
	procdev = 0;
    else
	procdev = sb.st_dev;
#endif

    /* check filetype */
    if(!strcmp(opt->filename, "-")) { /* read data from stdin */
	/*
	 * njh@bandsman.co.uk: treat the input as a mailbox, the program
	 * can then be used as a filter called when mail is received
	 */
	if(optl(opt, "mbox"))
	    ret = mbox(opt, trie, limits, threads);
	else
	    ret = checkstdin(trie, limits);

    } else if(strlen(opt->filename) == 0) {

	/* we need full path for some reasons (eg. archive handling) */
	if(!getcwd(cwd, 200)) {
	    mprintf("@Can't get absolute pathname of current working directory.\n");
	    ret = 57;
	} else
	    ret = scandirs(cwd, trie, threads, user, opt, limits);

    } else if((fmodeint = fileinfo(opt->filename, 2)) == -1) {
               mprintf("@Can't access file %s\n", opt->filename);
               perror(opt->filename);
               ret = 56;
    } else {
	fmode = (mode_t) fmodeint;

        if(compression && (opt->filename[0] != '/')) {
	    /* we need to complete the path */
            if(!getcwd(cwd, 200)) {
		mprintf("@Can't get absolute pathname of current working directory.\n");
                return 57;
            } else {
		fullpath = mcalloc(512, sizeof(char));
#ifdef NO_SNPRINTF
                sprintf(fullpath, "%s/%s", cwd, opt->filename);
#else
                snprintf(fullpath, 512, "%s/%s", cwd, opt->filename);
#endif
                mprintf("*Full path: %s\n", fullpath);
            }
        } else
	    fullpath = (char *) opt->filename;

        switch(fmode & S_IFMT) {
	    case S_IFREG:
		ret = scanfile(fullpath, trie, threads, user, opt, limits);
                break;

            case S_IFDIR:
		ret = scandirs(fullpath, trie, threads, user, opt, limits);
                break;

            default:
                mprintf("@Not supported file type (%s)\n", opt->filename);
                ret = 52;
	}

	if(compression && fullpath) {
	    free(fullpath);
	    fullpath = NULL;
	}
    }


    /* free the trie */
    cl_freetrie(trie);

    /* overwrite return code */
    if(claminfo.ifiles)
	ret = 1;
    else if(ret < 50) /* hopefully no error detected */ 
	ret = 0; /* just make sure it's 0 */

    return ret;
}

int scanfile(const char *filename, struct cl_node *root, int thread, const struct passwd *user, const struct optstruct *opt, const struct cl_limits *limits)
{
	int ret, options = 0, included;
	struct optnode *optnode;
	char *argument;
#ifdef C_LINUX
	struct stat sb;

    /* argh, don't scan /proc files */
    if(procdev)
	if(stat(filename, &sb) != -1)
	    if(sb.st_dev == procdev) {
		mprintf("%s: Excluded (/proc).\n", filename);
		return 0;
	    }
#endif

    if(optl(opt, "exclude")) {
	argument = getfirstargl(opt, "exclude",&optnode);
	while (argument) {
	    if(strstr(filename, argument)) {
		if(!printinfected)
		    mprintf("%s: Excluded.\n", filename);
		return 0;
	    }
	    argument = getnextargl(&optnode, "exclude");
	}
    }

    if(optl(opt, "include")) {
	included = 0;
	argument = getfirstargl(opt, "include",&optnode);
	while (argument && !included) {
	    if(strstr(filename, argument))
		included = 1;
	    argument = getnextargl(&optnode, "include");
	}

	if (!included) {
	    if(!printinfected)
		mprintf("%s: Excluded.\n", filename);
	    return 0;
	}
	
    }


    if(fileinfo(filename, 1) == 0) {
	if(!printinfected)
	    mprintf("%s: Empty file.\n", filename);
	return 0;
    }

    if(optl(opt, "disable-archive"))
	options &= ~CL_ARCHIVE;
    else
	options |= CL_ARCHIVE;

    /* check the extension */

    if((strbcasestr(filename, ".zip") || strbcasestr(filename, ".rar")) && (options & CL_ARCHIVE)) {
	/* try to use internal archivers */
	if((ret = checkfile(filename, root, limits, CL_ARCHIVE)) == CL_VIRUS)
	    return 1;
	else if(ret == CL_CLEAN)
	    return 0;
	/* in other case try to continue with external archivers */
	options &= ~CL_ARCHIVE; /* and disable decompression for the below checkfile() */
    }

    if((strbcasestr(filename, ".zip") && optl(opt, "unzip"))
    || (strbcasestr(filename, ".rar") && optl(opt, "unrar"))
    || (strbcasestr(filename, ".ace") && optl(opt, "unace"))
    || (strbcasestr(filename, ".arj") && optl(opt, "unarj"))
    || (strbcasestr(filename, ".zoo") && optl(opt, "zoo"))
    || (strbcasestr(filename, ".jar") && optl(opt, "jar"))
    || (strbcasestr(filename, ".lzh") && optl(opt, "lha"))
    || (strbcasestr(filename, ".tar") && optl(opt, "tar"))
    || (strbcasestr(filename, ".deb") && optl(opt, "deb"))
    || ((strbcasestr(filename, ".tar.gz") || strbcasestr(filename, ".tgz")) 
	 && (optl(opt, "tgz") || optl(opt, "deb"))) ) {

	/* check permissions */
	switch(readaccess(filename, UNPUSER)) {
	    case -1:
		mprintf("@Can't get information about user "UNPUSER".\n");
		texit(60); /* this is critical problem, so we just exit here */
	    case -2:
		mprintf("@Can't get information about current user.\n");
		texit(59); /* this is critical problem, so we just exit here */
	    case 0: /* read access denied */
		if(getuid()) {
		    if(!printinfected)
			mprintf("%s: Access denied to archive.\n", filename);
		} else {

		    if(limits && limits->maxfilesize)
			if(fileinfo(filename, 1) / 1024 > limits->maxfilesize) {
			    if(!printinfected)
				mprintf("%s: Archive too big.\n", filename);
			    return 0;
			}

		    return(scandenied(filename, root, thread, user, opt, limits));
		}
		return 0;
	    case 1:
		return(scancompressed(filename, root, thread, user, opt, limits));
	}
    }

    if(getuid())
	switch(readaccess(filename, NULL)) {
	    case -2:
		mprintf("@Can't get information about current user.\n");
		texit(59); /* this is critical problem, so we just exit here */
	    case 0: /* read access denied */
		if(!printinfected)
		    mprintf("%s: Access denied.\n", filename);
		return 0;
	}

    if((ret = checkfile(filename, root, limits, options)) == CL_VIRUS) {
	if(optl(opt, "remove")) {
	    if(unlink(filename)) {
		mprintf("%s: Can't remove.\n", filename);
		logg("%s: Can't remove.\n", filename);
		claminfo.notremoved++;
	    } else {
		mprintf("%s: Removed.\n", filename);
		logg("%s: Removed.\n", filename);
	    }
	}
    }
    return ret;
}

/* it has guaranteed read access to the archive */
int scancompressed(const char *filename, struct cl_node *root, int thread, const struct passwd *user, const struct optstruct *opt, const struct cl_limits *limits)
{
	int ret = 0;
	char *tmpdir, *gendir, *userprg;
	struct stat statbuf;

    stat(filename, &statbuf);

    if(!S_ISREG(statbuf.st_mode)) {
	mprintf("^Suspected archive %s is not a regular file.\n", filename);
	return 0; /* hmm ? */
    }

    /* check write access */

    if((tmpdir = getargl(opt, "tempdir")) == NULL)
	/* njh@bandsman.co.uk: use TMPDIR as an alternative */
	tmpdir = getenv("TMPDIR");

    if(tmpdir == NULL)
#ifdef P_tmpdir
	tmpdir = P_tmpdir;
#else
	tmpdir = "/tmp";
#endif

    if(writeaccess(tmpdir, UNPUSER) != 1) {
	mprintf("@Can't write to the temporary directory.\n");
	texit(64);
    }

    /* generate the temporary directory */

    gendir = gentemp(tmpdir);
    if(mkdir(gendir, 0700)) {
	mprintf("@Can't create the temporary directory %s\n", gendir);
	texit(63); /* critical */
    }

    if(user)
	chown(gendir, user->pw_uid, user->pw_gid);


    /* unpack file  - as unprivileged user */
    if(strbcasestr(filename, ".zip")) {
	char *args[] = { "unzip", "-P", "clam", "-o", (char *) filename, NULL };

	if((userprg = getargl(opt, "unzip")))
	    ret = clamav_unpack(userprg, args, gendir, user, opt);
	else
	    ret = clamav_unpack("unzip", args, gendir, user, opt);

    } else if(strbcasestr(filename, ".rar")) { 
	char *args[] = { "unrar", "x", "-p-", "-y", (char *) filename, NULL };
	if((userprg = getargl(opt, "unrar")))
	    ret = clamav_unpack(userprg, args, gendir, user, opt);
	else
	    ret = clamav_unpack("unrar", args, gendir, user, opt);

    } else if(strbcasestr(filename, ".ace")) { 
	char *args[] = { "unace", "x", "-y", (char *) filename, NULL };
	if((userprg = getargl(opt, "unace")))
	    ret = clamav_unpack(userprg, args, gendir, user, opt);
	else
	    ret = clamav_unpack("unace", args, gendir, user, opt);

    } else if(strbcasestr(filename, ".arj")) { 
	char *args[] = { "unarj", "x", (char *) filename, NULL };
	if((userprg = getargl(opt, "unarj")))
	    ret = clamav_unpack(userprg, args, gendir, user, opt);
	else
	    ret = clamav_unpack("unarj", args, gendir, user, opt);

    } else if(strbcasestr(filename, ".zoo")) { 
	char *args[] = { "zoo", "x.O", (char *) filename, NULL };
	if((userprg = getargl(opt, "zoo")))
	    ret = clamav_unpack(userprg, args, gendir, user, opt);
	else
	    ret = clamav_unpack("zoo", args, gendir, user, opt);

    } else if(strbcasestr(filename, ".jar")) { 
	char *args[] = { "unzip", "-P", "clam", "-o", (char *) filename, NULL };
	if((userprg = getargl(opt, "jar")))
	    ret = clamav_unpack(userprg, args, gendir, user, opt);
	else
	    ret = clamav_unpack("unzip", args, gendir, user, opt);

    } else if(strbcasestr(filename, ".lzh")) { 
	char *args[] = { "lha", "xf", (char *) filename, NULL };
	if((userprg = getargl(opt, "lha")))
	    ret = clamav_unpack(userprg, args, gendir, user, opt);
	else
	    ret = clamav_unpack("lha", args, gendir, user, opt);

    } else if(strbcasestr(filename, ".tar")) { 
	char *args[] = { "tar", "-xpvf", (char *) filename, NULL };
	if((userprg = getargl(opt, "tar")))
	    ret = clamav_unpack(userprg, args, gendir, user, opt);
	else
	    ret = clamav_unpack("tar", args, gendir, user, opt);

    } else if(strbcasestr(filename, ".deb")) { 
	char *args[] = { "ar", "x", (char *) filename, NULL };
	if((userprg = getargl(opt, "deb")))
	    ret = clamav_unpack(userprg, args, gendir, user, opt);
	else
	    ret = clamav_unpack("ar", args, gendir, user, opt);

    } else if((strbcasestr(filename, ".tar.gz") || strbcasestr(filename, ".tgz"))) {
	char *args[] = { "tar", "-zxpvf", (char *) filename, NULL };
	if((userprg = getargl(opt, "tgz")))
	    ret = clamav_unpack(userprg, args, gendir, user, opt);
	else
	    ret = clamav_unpack("tar", args, gendir, user, opt);
    }

    /* fix permissions of extracted files */
    fixperms(gendir);

    if(!ret) /* execute successful */
	/* notice, that we can't use threadscan() so also scandirs() here */
	ret = treewalk(gendir, root, thread, user, opt, limits);

    /* remove the directory  - as clamav */
    clamav_rmdirs(gendir);

    /* free gendir - it's not necessary now */
    free(gendir);

    switch(ret) {
	case -1:
	    mprintf("@Can't fork().\n");
	    texit(61); /* this is critical problem, so we just exit here */
	case -2:
	    mprintf("@Can't execute some unpacker. Check paths and permissions on the temporary directory.\n");
	    /* This is no longer a critical error (since 0.24). We scan
	     * raw archive.
	     */
	    if(!printinfected)
		mprintf("(raw) ");

	    if((ret = checkfile(filename, root, limits, 0)) == CL_VIRUS) {
		if(optl(opt, "remove")) {
		    if(unlink(filename)) {
			mprintf("%s: Can't remove.\n", filename);
			logg("%s: Can't remove.\n", filename);
			claminfo.notremoved++;
		    } else {
			mprintf("%s: Removed.\n", filename);
			logg("%s: Removed.\n", filename);
		    }
		}
	    }
	    return ret;
	case -3:
	    return 0;
	case 0:
	    /* no viruses found in archive, we scan just in case the same
	     * archive
	     */
	    if(!printinfected)
		mprintf("(raw) ");

	    if((ret = checkfile(filename, root, limits, 0)) == CL_VIRUS) {
		if(optl(opt, "remove")) {
		    if(unlink(filename)) {
			mprintf("%s: Can't remove.\n", filename);
			logg("%s: Can't remove.\n", filename);
			claminfo.notremoved++;
		    } else {
			mprintf("%s: Removed.\n", filename);
			logg("%s: Removed.\n", filename);
		    }
		}
	    }
	    return ret;
	case 1:
	    logg("%s: Infected Archive FOUND\n", filename);
	    mprintf("%s: Infected Archive FOUND\n", filename);
	    if(optl(opt, "remove")) {
		if(unlink(filename)) {
		    mprintf("%s: Can't remove.\n", filename);
		    logg("%s: Can't remove.\n", filename);
		    claminfo.notremoved++;
		} else {
		    mprintf("%s: Removed.\n", filename);
		    logg("%s: Removed.\n", filename);
		}
	    }
	    return 1;
	default:
	    mprintf("@Strange value (%d) returned in scancompressed()\n", ret);
	    return 0;
    }
}

int scandenied(const char *filename, struct cl_node *root, int thread, const struct passwd *user, const struct optstruct *opt, const struct cl_limits *limits)
{
	char *tmpdir, *gendir, *tmpfile, *pt;
	struct stat statbuf;
	int ret;

    stat(filename, &statbuf);
    if(!S_ISREG(statbuf.st_mode)) {
	mprintf("^Suspected archive %s is not a regular file.\n", filename);
	return 0;
    }

    /* check write access */

    if((tmpdir = getargl(opt, "tempdir")) == NULL)
        tmpdir = getenv("TMPDIR");

    if(tmpdir == NULL)
#ifdef P_tmpdir
	tmpdir = P_tmpdir;
#else
	tmpdir = "/tmp";
#endif


    if(writeaccess(tmpdir, UNPUSER) != 1) {
	mprintf("@Can't write to the temporary directory %s.\n", tmpdir);
	texit(64);
    }

    /* generate the temporary directory */
    gendir = gentemp(tmpdir);
    if(mkdir(gendir, 0700)) {
	mprintf("@Can't create the temporary directory %s\n", gendir);
	texit(63); /* critical */
    }

    tmpfile = (char *) mcalloc(strlen(gendir) + strlen(filename) + 10, sizeof(char));
    pt = strrchr(filename, '/');
    if(!pt)
	pt = (char *) filename;
    else
	pt += 1;

    sprintf(tmpfile, "%s/%s", gendir, pt);

    if(filecopy(filename, tmpfile) == -1) {
	mprintf("!I/O error.\n");
	perror("copyfile()");
	texit(58);
    }

    fixperms(gendir);

    if(user) {
	chown(gendir, user->pw_uid, user->pw_gid);
	chown(tmpfile, user->pw_uid, user->pw_gid);
    }

    if((ret = treewalk(gendir, root, thread, user, opt, limits)) == 1) {
	logg("(Real infected archive: %s)\n", filename);
	mprintf("(Real infected archive: %s)\n", filename);

	if(optl(opt, "remove")) {
	    if(unlink(filename)) {
		mprintf("%s: Can't remove.\n", filename);
		logg("%s: Can't remove.\n", filename);
		claminfo.notremoved++;
	    } else {
	        mprintf("%s: Removed.\n", filename);
	        logg("%s: Removed.\n", filename);
	    }
        }
    }

    /* remove the directory  - as clamav */
    clamav_rmdirs(gendir);

    free(gendir);
    free(tmpfile);

    return ret;
}

int scandirs(const char *dirname, struct cl_node *root, int thread, const struct passwd *user, const struct optstruct *opt, const struct cl_limits *limits)
{

#ifdef CLAMSCAN_THREADS
    if(thread)
	return(threadscan(dirname, root, thread, user, opt, limits));
    else
#endif
	return(treewalk(dirname, root, thread, user, opt, limits));
}

int checkfile(const char *filename, const struct cl_node *root, const struct cl_limits *limits, int options)
{
	int fd, ret;
	char *virname;

    if((fd = open(filename, O_RDONLY)) == -1) {
	mprintf("@Can't open file %s\n", filename);
	return 54;
    }

    claminfo.files++;

    if((ret = cl_scandesc(fd, &virname, &claminfo.blocks, root, limits, options)) == CL_VIRUS) {
	mprintf("%s: %s FOUND\n", filename, virname);
	claminfo.ifiles++;
    } else if(ret == CL_CLEAN) {
	if(!printinfected)
	    mprintf("%s: OK\n", filename);
    } else
	if(!printinfected)
	    mprintf("%s: %s\n", filename, cl_perror(ret));

    close(fd);
    return ret;
}

int checkstdin(const struct cl_node *root, const struct cl_limits *limits)
{
	int ret;
	char *virname;


    claminfo.files++;

    if((ret = cl_scandesc(0, &virname, &claminfo.blocks, root, limits, CL_RAW)) == CL_VIRUS) {
	mprintf("stdin: %s FOUND\n", virname);
	claminfo.ifiles++;
    } else if(ret == CL_CLEAN) {
	if(!printinfected)
	    mprintf("stdin: OK\n");
    } else
	if(!printinfected)
	    mprintf("stdin: %s\n", cl_perror(ret));

    return ret;
}

#ifdef CLAMSCAN_THREADS
void thr_exit(int sig)
{
	int i;

    /* it's brutal, but signal handling is not yet implemented */
    for(i = 0; i < thr_threads; i++)
	pthread_kill(thr_threadtab[i], 9);

    exit(thr_exitno);
}
#endif

/*
 * -1 -> can't fork
 * -2 -> can't execute
 * -3 -> external signal
 * 0 -> OK
 */

int clamav_unpack(const char *prog, char **args, const char *tmpdir, const struct passwd *user, const struct optstruct *opt)
{
	pid_t pid;
	int status, wret, maxfiles, maxspace;
	struct s_du n;


    if(optl(opt, "max-files"))
	maxfiles = atoi(getargl(opt, "max-files"));
    else
	maxfiles = 0;

    if(optl(opt, "max-space")) {
	    char *cpy, *ptr;
	ptr = getargl(opt, "max-space");
	if(tolower(ptr[strlen(ptr) - 1]) == 'm') { /* megabytes */
	    cpy = mcalloc(strlen(ptr), sizeof(char));
	    strncpy(cpy, ptr, strlen(ptr) - 1);
	    maxspace = atoi(cpy) * 1024;
	    free(cpy);
	} else /* default - kilobytes */
	    maxspace = atoi(ptr);
    } else
	maxspace = 0;


    switch(pid = fork()) {
	case -1:
	    return -1;
	case 0:
#ifndef C_CYGWIN
	    if(!getuid() && user) {
		setgid(user->pw_gid);
		setuid(user->pw_uid);
	    }
#endif
	    chdir(tmpdir);

	    if(printinfected) {
		close(1);
		close(2);
	    }

	    if(strchr(prog, '/')) /* we have full path */
		execv(prog, args);
	    else
		execvp(prog, args);
	    perror("execv(p)");
	    abort();
	    break;
	default:

	    if(maxfiles || maxspace) {
		while(!(wret = waitpid(pid, &status, WNOHANG))) {
		    memset(&n, 0, sizeof(struct s_du));

		    if(!du(tmpdir, &n))
			if((maxfiles && n.files > maxfiles) || (maxspace && n.space > maxspace)) {
			    mprintf("*n.files: %d, n.space: %d\n", n.files, n.space);
			    kill(pid, 9); /* stop it immediately */
			}
		}
	    } else
		waitpid(pid, &status, 0);


	    if(WIFSIGNALED(status)) {
		switch(WTERMSIG(status)) {

		    case 9:
			mprintf("\nUnpacker process %d stopped due to exceeded limits.\n", pid);
			return 0;
		    case 6: /* abort */
			mprintf("@Can't run %s\n", prog);
			return -2;
		    default:
			mprintf("@\nUnpacker stopped with external signal %d\n", WTERMSIG(status));
			return -3;
		}
	    } else if(WIFEXITED(status))
		return 0;
    }

    return 0;
}

