/*
 *  Copyright (C) 2002 Nigel Horne <njh@bandsman.co.uk>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef	CL_DEBUG
#define	NDEBUG	/* map CLAMAV debug onto standard */
#endif

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <assert.h>
#include <string.h>
#include <strings.h>
#include <ctype.h>
#include <time.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <clamav.h>

#include "table.h"
#include "blob.h"
#include "mbox.h"
#include "text.h"
#include "message.h"
#include "others.h"
#include "defaults.h"
#include "treewalk.h"

typedef enum    { FALSE = 0, TRUE = 1 } bool;

static	int	insert(const message *mainMessage, blob **blobsIn, int nBlobs, text *textIn, const char *tmpdir);
int	boundaryStart(const char *line, const char *boundary);
int	endOfMessage(const char *line, const char *boundary);
static	int	initialiseTables(void);
static	int	getTextPart(message *const messages[], size_t size);
static	size_t	strip(char *buf, int len);
static	size_t	strstrip(char *s);
static	bool	continuationMarker(const char *line);


/* Maximum number of attachements that we accept */
#define	MAX_ATTACHMENTS	10

/* Maximum line length according to RFC821 */
#define	LINE_LENGTH	1000

/* Hashcodes for our hash tables */
#define	CONTENT_TYPE			1
#define	CONTENT_TRANSFER_ENCODING	2
#define	CONTENT_DISPOSITION		3

/* Mime sub types */
#define	PLAIN		1
#define	ENRICHED	2
#define	HTML		3
#define	RICHTEXT	4
#define	MIXED		5
#define	ALTERNATIVE	6
#define	DIGEST		7
#define	SIGNED		8
#define	PARALLEL	9
#define	RELATED		10	/* RFC2387 */

static	struct tableinit {
	const	char	*key;
	int	value;
} rfc821headers[] = {
	{	"Content-Type:",		CONTENT_TYPE		},
	{	"Content-Transfer-Encoding:",	CONTENT_TRANSFER_ENCODING	},
	{	"Content-Disposition:",		CONTENT_DISPOSITION	},
	{	NULL,				0			}
}, mimeSubtypes[] = {
		/* subtypes of Text */
	{	"plain",	PLAIN		},
	{	"enriched",	ENRICHED	},
	{	"html",		HTML		},
	{	"richtext",	RICHTEXT	},
		/* subtypes of Multipart */
	{	"mixed",	MIXED		},
	{	"alternative",	ALTERNATIVE	},
	{	"digest",	DIGEST		},
	{	"signed",	SIGNED		},
	{	"parallel",	PARALLEL	},
	{	"related",	RELATED		},
	{	NULL,		0		}
};

static	table_t	*rfc821Table, *subtypeTable;

int
mbox(const struct optstruct *opt, struct cl_node *root, struct cl_limits *limits, int threads)
{
	int retcode;
	char buffer[LINE_LENGTH];
	bool first = TRUE;
	bool inHeader = FALSE;
	bool inMimeHeader = FALSE;
	bool lastLineWasEmpty = TRUE;
	message *m;
	char *tmpdir;
	struct passwd *user = NULL;

	mprintf("*in mbox()\n");

	if(initialiseTables() < 0)
		return -1;

	retcode = 0;

	m = messageCreate();
	assert(m != NULL);

	/*
	 * a lot of code here duplicated with scanmanager - need to clean it up
	 */

	/* create the directory to extract the files */

	if((tmpdir = getargl(opt, "tempdir")) == NULL) {
		tmpdir = getenv("TMPDIR");
		if(tmpdir == NULL)
#ifdef	P_tmpdir
			tmpdir = P_tmpdir;
#else
			tmpdir = "/tmp";
#endif
	}

	if(writeaccess(tmpdir, UNPUSER) != 1) {
		mprintf("@Can't write to the temporary directory.\n");
		texit(64);
	}

	/* generate the temporary directory */

	tmpdir = gentemp(tmpdir);
	if(mkdir(tmpdir, 0700)) {
		mprintf("@Can't create the temporary directory %s\n", tmpdir);
		texit(63); /* critical */
	}

	/* njh@bandsman.co.uk: BeOS */
#if !defined(C_CYGWIN) && !defined(C_BEOS)
    if(!getuid()) {
	if((user = getpwnam(UNPUSER)) == NULL) {
	    mprintf("@Can't get information about user "UNPUSER".\n");
	    exit(60); /* this is critical problem, so we just exit here */
	}
    }
#endif

	if(user)
		chown(tmpdir, user->pw_uid, user->pw_gid);

	/*
	 * handle more than one message in the filter. Probably a waste of time
	 */
	while(fgets(buffer, sizeof(buffer), stdin) != NULL) {
		if(first)
			/* sanity check */
			if(strncmp(buffer, "From ", 5) != 0) {
				mprintf("!Not a valid mail message");
				retcode = -1;
				break;
			}

		/*
		 * Handle this where we're mid point through this stuff
		 *	Content-Type: multipart/alternative;
		 *		boundary="----foo"
		 */
		if(inMimeHeader) {
			const char *ptr;

			assert(!first);

			if(!continuationMarker(buffer))
				inMimeHeader = FALSE;	 /* no more args */

			/*
			 * Add all the arguments on the line
			 */
			for(ptr = strtok(buffer, ";\r\n"); ptr; ptr = strtok(NULL, ":\r\n"))
				messageAddArgument(m, ptr);

		} else if((!inHeader) && lastLineWasEmpty && (strncmp(buffer, "From ", 5) == 0)) {
			/*
			 * New message, save the previous message, if any
			 */
			if(!first) {
				/*
				 * End of the current message, add it and look
				 * for the start of the next one
				 */
				messageClean(m);
				if(messageGetBody(m)) {
					if(!insert(m,  NULL, 0, NULL, tmpdir))
						break;
				}
				messageDestroy(m);
				m = messageCreate();
				assert(m != NULL);
			} else
				first = FALSE;

			lastLineWasEmpty = inHeader = TRUE;
		} else if(inHeader) {
			/*
			 * A blank line signifies the end of the header and
			 * the start of the text
			 */
			strstrip(buffer);

			if((buffer[0] == '\n') || (buffer[0] == '\r')) {
				mprintf("*End of header information\n");
				inHeader = 0;
			} else {
				const bool isLastLine = !continuationMarker(buffer);
				const char *cmd = strtok(buffer, " \t");
				char *arg = strtok(NULL, "\r\n");

				switch(tableFind(rfc821Table, cmd)) {
				case CONTENT_TYPE:
					/*
					 * Fix for non RFC1521 compliant mailers
					 * that send content-type: Text instead
					 * of content-type: Text/Plain, or
					 * just simply "Content-Type:"
					 */
					if(arg == NULL)
						  mprintf("*Empty content-type received, no subtype specified, assuming text/plain; charset=us-ascii\n");
					else if(strchr(arg, '/') == NULL)
						  mprintf("*Invalid content-type '%s' received, no subtype specified, assuming text/plain; charset=us-ascii\n", arg);
					else {
						/*
						 * Some clients are broken and
						 * put white space after the ;
						 */
						strstrip(arg);
						mprintf("*content-type: parse line '%s'\n", arg);
						messageSetMimeType(m, strtok(arg, "/"));

						messageSetMimeSubtype(m, strtok(NULL, ";"));

						/*
						 * Add in all the arguments.
						 */
						while((arg = strtok(NULL, "\r\n \t")))
							messageAddArgument(m, arg);
					}

					inMimeHeader = !isLastLine;
					break;
				case CONTENT_TRANSFER_ENCODING:
					messageSetEncoding(m, arg);
					break;
				}
			}
		} else {
			assert(!first);

			lastLineWasEmpty = ((buffer[0] == '\n') || (buffer[0] == '\r'));
			/*
			 * Add this line to the end of the linked list
			 * of lines. This isn't needed when using
			 * .forward since the rest of the file *must*
			 * be the text so a single fread() should
			 * suffice. Still, it does no harm and is more
			 * flexible this way
			 *
			 * Note that the terminating newline is not
			 * added
			 */
			messageAddLine(m, strtok(buffer, "\r\n"));
		}
	}

	/*
	 * Write out the last entry in the mailbox
	 */
	if(retcode == 0) {
		messageClean(m);
		if(messageGetBody(m))
			if(!insert(m, NULL, 0, NULL, tmpdir))
				retcode = -1;
	}

	/*
	 * Tidy up and quit
	 */
	messageDestroy(m);

	tableDestroy(rfc821Table);
	tableDestroy(subtypeTable);

	/* fix permissions of extracted files */
	fixperms(tmpdir);

	if(retcode == 0) /* execute successful */
		retcode = treewalk(tmpdir, root, threads, user, opt, limits);

	/* remove the directory - as clamav */
	clamav_rmdirs(tmpdir);

	/* free tmpdir - it's not necessary now */
	free(tmpdir);

	return retcode;
}

/*
 * This is a recursive routine.
 *
 * mainMessage is the buffer to be parsed. First time of calling it'll be
 *	the whole message. Later it'll be parts of a multipart message
 * textIn is the plain text message being built up so far
 * blobsIn contains the array of attachments found so far
 */
static int	/* success or fail */
insert(const message *mainMessage, blob **blobsIn, int nBlobs, text *textIn, const char *tmpdir)
{
	char *ptr;
	const char *boundary;
	message *messages[MAXALTERNATIVE], *aMessage;
	int multiparts, inhead, inMimeHead, i, rc;
	text *aText;
	blob *blobList[MAX_ATTACHMENTS], **blobs;
	short attachmentNumber;

	mprintf("*in insert()\n");

	/* Pre-assertions */
	assert(nBlobs < MAX_ATTACHMENTS);

	aText = textIn;
	blobs = blobsIn;

	/* Anything left to be parsed? */
	if(mainMessage) {
		int numberOfAttachments = 0;
		int plaintextPart;
		mime_type mimeType;
		const char *mimeSubtype;
		const text *t_line;

		assert(messageGetBody(mainMessage) != NULL);
		mimeType = messageGetMimeType(mainMessage);
		mimeSubtype = messageGetMimeSubtype(mainMessage);

		if((mimeType == TEXT) && (tableFind(subtypeTable, mimeSubtype) == PLAIN)) {
			/*
			 * This is effectively no encoding, notice that we
			 * don't check that charset is us-ascii
			 */
			mprintf("*assume no encoding\n");
			mimeType = NOMIME;
		}

		switch(mimeType) {
		case NOMIME:
			aText = textAddMessage(aText, mainMessage);
			break;
		case TEXT:
			if(tableFind(subtypeTable, mimeSubtype) == PLAIN)
				aText = textCopy(messageGetBody(mainMessage));
			break;
		case MULTIPART:
			assert(mimeSubtype[0] != '\0');

			boundary = messageFindArgument(mainMessage, "boundary");

			assert(boundary != NULL);

			/*
			 * Get to the start of the first message
			 */
			for(t_line = messageGetBody(mainMessage); t_line; t_line = t_line->t_next)
				if(boundaryStart(t_line->t_text, boundary))
					break;

			if(t_line == NULL) {
				mprintf("^Multipart MIME message contains no parts");
				return 1;	/* Nothing to do */
			}
			/*
			 * Build up a table of all of the parts of this
			 * multipart message. Remember, each part may itself
			 * be a multipart message.
			 */
			inhead = 1;
			inMimeHead = 0;

			for(multiparts = 0; t_line && (multiparts < MAXALTERNATIVE); multiparts++) {

				aMessage = messages[multiparts] = messageCreate();

				/* tk: shut up parentheses warning */
				while((t_line = t_line->t_next)) {
					const char *line = t_line->t_text;

					if(inMimeHead) {
						while(isspace(*line))
							line++;

						if(*line == '\0') {
							inhead = inMimeHead = 0;
							continue;
						}
						/*
						 * This may cause a trailing ';'
						 * to be added if this test
						 * fails - TODO: verify this
						 */
						inMimeHead = continuationMarker(line);
						messageAddArgument(aMessage, line);
					} else if(inhead) {
						char *copy;

						if(strlen(line) == 0) {
							inhead = 0;
							continue;
						}
						/*
						 * Some clients are broken and
						 * put white space after the ;
						 */
						inMimeHead = continuationMarker(line);
						copy = strdup(line);
						ptr = strtok(copy, " \t");


						switch(tableFind(rfc821Table, ptr)) {
						case CONTENT_TYPE:
							messageSetMimeType(aMessage, strtok(NULL, "/"));

							messageSetMimeSubtype(aMessage, strtok(NULL, ";"));

							messageAddArgument(aMessage, strtok(NULL, ";"));
							break;
						case CONTENT_TRANSFER_ENCODING:
							messageSetEncoding(aMessage, strtok(NULL, ""));
							break;
						case CONTENT_DISPOSITION:
							messageSetDispositionType(aMessage, strtok(NULL, ";"));
							messageAddArgument(aMessage, strtok(NULL, ""));
							break;
						}
						free(copy);
					} else if(boundaryStart(line, boundary)) {
						inhead = 1;
						break;
					} else if(endOfMessage(line, boundary)) {
						t_line = NULL;
						break;
					} else
						messageAddLine(aMessage, line);
				}
				messageClean(aMessage);
			}

			free((char *)boundary);


			if(multiparts == 0)
				return 1;	/* Nothing to do */


			switch(tableFind(subtypeTable, mimeSubtype)) {
			case ALTERNATIVE:
				plaintextPart = getTextPart(messages, multiparts);

				if(plaintextPart == -1) {
					plaintextPart = 0;
				}

				aMessage = messages[plaintextPart];
				aText = textAddMessage(aText, aMessage);

				rc = insert(NULL, blobs, nBlobs, aText, tmpdir);
				blobArrayDestroy(blobs, nBlobs);
				/*
				 * Fall through - some clients are broken and
				 * say alternative instead of mixed. The Klez
				 * virus is broken that way
				 */
			case MIXED:
				/*
				 * Look for attachments
				 *
				 * Not all formats are supported. If an
				 * unsupported format turns out to be
				 * common enough to implement, it is a simple
				 * matter to add it
				 */
				if(aText)
					mainMessage = NULL;

				for(i = 0; i < multiparts; i++) {
					const char *dtype;
					bool addAttachment = FALSE;
					bool addToText = FALSE;

					aMessage = messages[i];

					assert(aMessage != NULL);

					dtype = messageGetDispositionType(aMessage);

					switch(messageGetMimeType(aMessage)) {
					case APPLICATION:
						/*
						 * We don't care about the application
						 * subtype, since we don't spawn
						 * anything off
						 */
						if((strcasecmp(dtype, "attachment") == 0) ||
						   (strcasecmp(dtype, "octet-stream")))
							addAttachment = TRUE;
						else {
							mprintf("*Discarded application not sent as attachment\n");
							continue;
						}

						break;
					case NOMIME:
						mainMessage = NULL;
						addToText = TRUE;
						if(messageGetBody(aMessage) == NULL)
							/*
							 * No plain text version
							 */
							messageAddLine(aMessage, "No plain text alternative");
						assert(messageGetBody(aMessage) != NULL);
						break;
					case TEXT:
						if(strcasecmp(dtype, "attachment") == 0)
							addAttachment = TRUE;
						else if((*dtype == '\0') || (strcasecmp(dtype, "inline") == 0)) {
							mainMessage = NULL;
							addToText = TRUE;
						} else {
							mprintf("^Text type %s is not supported", dtype);
							continue;
						}
						break;
					case MULTIPART:
						/*
						 * It's a multi part within a multi part
						 * Run the message parser on this bit, it won't
						 * be an attachment
						 *
						 */

						mainMessage = aMessage;
						continue;
					case AUDIO:
					case IMAGE:
						/*
						 * TODO: it may be nice to
						 * have an option to throw
						 * away all images and sound
						 * files for ultra-secure sites
						 */
						addAttachment = TRUE;
						break;
					default:
						mprintf("^Only text and application attachments are supported, type = %d\n",
							messageGetMimeType(aMessage));
						continue;
					}

					/*
					 * It must be either text or
					 * an attachment. It can't be both
					 */
					assert(addToText || addAttachment);
					assert(!(addToText && addAttachment));

					if(addToText)
						aText = textAdd(aText, messageGetBody(aMessage));
					if(addAttachment) {
						blob *aBlob = messageToBlob(aMessage, NULL);

						if(aBlob) {
							assert(blobGetFilename(aBlob) != NULL);
							/*if(blobGetDataSize(aBlob) > 0)*/
								blobList[numberOfAttachments++] = aBlob;
						}
					}
				}

				if(numberOfAttachments == 0) {
					/* No usable attachment was found */
					rc = insert(NULL, NULL, 0, aText, tmpdir);
					break;
				}
				/*
				 * Store any existing attachments at the end of
				 * the list we've just built up
				 */
				for(i = 0; i < nBlobs; i++)
					blobList[numberOfAttachments++] = blobs[i];
				rc = insert(mainMessage, blobList, numberOfAttachments, aText, tmpdir);
				break;
			case RELATED:
				/*
				 * Look for the text bit
				 */
				aMessage = NULL;
				assert(multiparts > 0);

				plaintextPart = getTextPart(messages, multiparts);

				if(plaintextPart >= 0)
					aText = textAddMessage(aText, messages[plaintextPart]);
				else {
					/*
					 * There isn't a text bit. If there's a
					 * multipart bit, it'll probably be in
					 * there somewhere
					 */
					for(i = 0; i < multiparts; i++) {
						if(messageGetMimeType(messages[i]) == MULTIPART) {
							aMessage = messages[i];
							plaintextPart = i;
							break;
						}
					}
				}

				assert(plaintextPart != -1);

				rc = insert(aMessage, blobs, nBlobs, aText, tmpdir);
				/*
				 * This causes a memory fault.
				 * I suspect that at this point two things are pointing
				 * to the same thing, this causing a multiple free
				 */
				/*blobArrayDestroy(blobs, nBlobs);*/
				break;
			case DIGEST:
			case SIGNED:
			case PARALLEL:
				/*
				 * If we're here it could be because we have a
				 * multipart/mixed message, consisting of a
				 * message followed by an attachment. That
				 * message itself is a multipart/alternative
				 * message and we need to dig out the plain
				 * text part of that alternative
				 */
				plaintextPart = getTextPart(messages, multiparts);
				if(plaintextPart == -1)
					plaintextPart = 0;

				rc = insert(messages[plaintextPart], blobs, nBlobs, aText, tmpdir);
				blobArrayDestroy(blobs, nBlobs);
				break;
			default:
				mprintf("^Unsupported multipart format `%s'\n", mimeSubtype);
				rc = 0;
			}

			for(i = 0; i < multiparts; i++)
				messageDestroy(messages[i]);

			if(blobs && (blobsIn == NULL))
				puts("arraydestroy");

			return rc;

		case MESSAGE:
			/*
			 * Check for forbidden encodings
			 */
			switch(messageGetEncoding(mainMessage)) {
				case NOENCODING:
				case EIGHTBIT:
				case BINARY:
					break;
				default:
					mprintf("^MIME type message can not be encoded");
					break;
			}
			if(strcasecmp(mimeSubtype, "rfc822") == 0)
				/*
				 * Just copy it through as it is for now.
				 * TODO: pick out the encapsulated message
				 */
				break;
			else if(strcasecmp(mimeSubtype, "partial") == 0)
				/* TODO */
				mprintf("^Content-type message/partial not yet supported");
			else if(strcasecmp(mimeSubtype, "external-body") == 0)
				/*
				 * I don't believe that we should be going
				 * around the Internet looking for referenced
				 * files...
				 */
				mprintf("^Attempt to send Content-type message/external-body trapped");
			else
				mprintf("^Unsupported message format `%s'", mimeSubtype);

			return 0;

		case VIDEO:
		case AUDIO:
		case IMAGE:
		case APPLICATION:
			break;

		default:
			mprintf("^Message received with unknown mime encoding");
			break;
		}
	}

	for(attachmentNumber = 0; attachmentNumber < nBlobs; attachmentNumber++) {
		const blob *b = blobs[attachmentNumber];
		unsigned long nbytes;
		int fd;
		const char *cptr;
		char *filename;

		assert(b != NULL);
		nbytes = blobGetDataSize(b);

		if(nbytes == 0)
			continue;
		/*
		 * message_id
		 * attachment number
		 * filename
		 * blob of attachment
		 */
		cptr = blobGetFilename(b);

		assert(cptr != NULL);

		/* tk: use dynamic allocation */
		filename = mcalloc(strlen(tmpdir) + strlen(cptr) + 10, sizeof(char));
		sprintf(filename, "%s/%sXXXXXX", tmpdir, cptr);

		/*
		 * TODO: add a HAS_MKSTEMP property
		 */
#if	defined(C_LINUX)
		fd = mkstemp(filename);
#else
		(void)mktemp(filename);
		fd = open(filename, O_WRONLY|O_CREAT|O_EXCL, 0600);
#endif

		if(fd < 0) {
			mprintf("@%s: %s\n", filename, strerror(errno));
			texit(64);
		} else {
			write(fd, blobGetData(b), nbytes);
			close(fd);
			mprintf("*Attachment saved as %s\n", filename);
			free(filename);
		}
	}

	if(aText && (textIn == NULL))
		textDestroy(aText);

	if(blobs && (blobsIn == NULL))
		puts("arraydestroy");

	return 1;
}


/*
 * Is the current line the start of a new section?
 *
 * New sections start with --boundary
 */
int
boundaryStart(const char *line, const char *boundary)
{
	if(*line++ != '-')
		return 0;
	if(*line++ != '-')
		return 0;
	return strcasecmp(line, boundary) == 0;
}

/*
 * Is the current line the end?
 *
 * The message ends with with --boundary--
 */
int
endOfMessage(const char *line, const char *boundary)
{
	size_t len;

	if(*line++ != '-')
		return 0;
	if(*line++ != '-')
		return 0;
	if(strncasecmp(line, boundary, strlen(boundary)) != 0)
		return 0;
	len = strlen(boundary);
	if(strlen(line) != (len + 2))
		return 0;
	line = &line[len];
	if(*line++ != '-')
		return 0;
	return *line == '-';
}

/*
 * Initialise the various lookup tables
 */
static int
initialiseTables(void)
{
	const struct tableinit *tableinit;

	/*
	 * Initialise the various look up tables
	 */
	rfc821Table = tableCreate();
	assert(rfc821Table != NULL);

	for(tableinit = rfc821headers; tableinit->key; tableinit++)
		if(tableInsert(rfc821Table, tableinit->key, tableinit->value) < 0)
			return -1;

	subtypeTable = tableCreate();
	assert(subtypeTable != NULL);

	for(tableinit = mimeSubtypes; tableinit->key; tableinit++)
		if(tableInsert(subtypeTable, tableinit->key, tableinit->value) < 0)
			return -1;

	return 0;
}

/*
 * If there's a plain text version use that, otherwise
 * use the first text part, otherwise just use the
 * first one around
 *
 * Alternatively we could hunt out any HTML
 * version and save that, if the displaying client
 * can cope with HTML.
 *
 * If we can't find one, return -1
 */
static int
getTextPart(message *const messages[], size_t size)
{
	int i;

	for(i = 0; i < size; i++) {
		assert(messages[i] != NULL);
		if((messageGetMimeType(messages[i]) == TEXT) &&
		   (strcasecmp(messageGetMimeSubtype(messages[i]), "plain") == 0))
			return i;
	}
	for(i = 0; i < size; i++)
		if(messageGetMimeType(messages[i]) == TEXT)
			return i;

	return -1;
}

/*
 * strip -
 *	Remove the trailing spaces from a buffer
 * Returns it's new length (a la strlen)
 *
 * len must be int not size_t because of the >= 0 test, it is sizeof(buf)
 *	not strlen(buf)
 */
static size_t
strip(char *buf, int len)
{
	register char *ptr;
	register size_t i;

	if((buf == NULL) || (len <= 0))
		return(0);

	i = strlen(buf);
	if(len > (int)(i + 1))
		return(i);

	ptr = &buf[--len];

#if	defined(UNIX) || defined(C_LINUX) || defined(C_DARWIN)	/* watch - it may be in shared text area */
	do
		if(*ptr)
			*ptr = '\0';
	while((--len >= 0) && !isgraph(*--ptr) && (*ptr != '\n') && (*ptr != '\r'));
#else	/* more characters can be displayed on DOS */
	do
#ifndef	REAL_MODE_DOS
		if(*ptr)	/* C8.0 puts into a text area */
#endif
			*ptr = '\0';
	while((--len >= 0) && (*--ptr == '\0') || (isspace(*ptr)));
#endif
	return((size_t)(len + 1));
}

/*
 * strstrip:
 *	Strip a given string
 */
static size_t
strstrip(char *s)
{
	if(s == (char *)NULL)
		return(0);
	return(strip(s, strlen(s) + 1));
}

/*
 * When parsing a MIME header see if this spans more than one line. A
 * semi-colon at the end of the line indicates that the MIME information
 * is continued on the next line.
 *
 * Some clients are broken and put white space after the ;
 */
static bool
continuationMarker(const char *line)
{
	const char *ptr;

	assert(line != NULL);

	/*printf("continuationMarker '%s'", line);*/

	ptr = strchr(line, '\0');

	assert(ptr != NULL);

	while(ptr >= line)
		switch(*--ptr) {
			case '\n':
			case '\r':
			case ' ':
			case '\t':
				continue;
			case ';':
				return TRUE;
			default:
				return FALSE;
		}

	return FALSE;
}
