/*
 *  Copyright (C) 2002 Tomasz Kojm <zolw@konarski.edu.pl>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <pwd.h>

#include "options.h"
#include "shared.h"
#include "others.h"
#include "manager.h"
#include "defaults.h"

void help(void);
void daemonize(void);

/* this local macro takes care about freeing memory at exit */
/*
#define mexit(i)    if(opt) free_opt(opt);			    \
		    mprintf("*Memory freed. Exit code: %d\n", i);   \
		    exit(i);
*/
#define mexit(i)    exit(i)

void freshclam(struct optstruct *opt)
{
	int ret;
#ifndef C_CYGWIN
	struct passwd *user;

    /* freshclam shouldn't work with root priviledges */
    if(!getuid()) { 
	if((user = getpwnam(UNPUSER)) == NULL) {
	    mprintf("@Can't get information about user "UNPUSER".\n");
	    exit(60); /* this is critical problem, so we just exit here */
	}

	setgroups(1, &user->pw_gid);
	setgid(user->pw_gid);
	setuid(user->pw_uid);
    }
#endif

    /* initialize some important variables */

    mprintf_disabled = 0;

    if(optc(opt, 'v')) mprintf_verbose = 1;
    else mprintf_verbose = 0;

    if(optl(opt, "quiet")) mprintf_quiet = 1;
    else mprintf_quiet = 0;

    if(optl(opt, "stdout")) mprintf_stdout = 1;
    else mprintf_stdout = 0;

    if(optc(opt, 'V')) {
	mprintf("freshclam / ClamAV version "VERSION"\n");
	mexit(0);
    }

    if(optc(opt, 'h')) {
	free_opt(opt);
    	help();
    }

    /* initialize logger */

    if(optl(opt, "log-verbose")) logverbose = 1;
    else logverbose = 0;

    if(optc(opt, 'l')) {
	logfile = getargc(opt, 'l');
	if(logg("--------------------------------------\n")) {
	    mprintf("!Problem with internal logger.\n");
	    mexit(1);
	}
    } else 
	logfile = NULL;


    if(optc(opt, 'd')) {
	    int bigsleep, checks;

	if(!optc(opt, 'c')) {
	    mprintf("@Daemon mode requires -c (--checks) option.\n");
	    mexit(40);
	}

	checks = atoi(getargc(opt, 'c'));

	if(checks <= 0 || checks > 50) {
	    mprintf("@Wrong number of checks\n");
	    mexit(41);
	}

	bigsleep = 24*3600 / checks;
	daemonize();

	while(1) {
	    downloadmanager(opt);
	    logg("\n--------------------------------------\n");
	    sleep(bigsleep);
	}

    } else
	ret = downloadmanager(opt);

    mexit(ret);

}

void daemonize(void)
{
	int i;

    for(i = 0; i < 3; i++)
	close(i);

    umask(0);

    if(fork())
	exit(0);

    setsid();
    mprintf_disabled = 1;
}

void help(void)
{

    mprintf_stdout = 1;

    mprintf("\n"
"		   Clam AntiVirus: FreshClam  "VERSION"\n"
"		   (c) 2002 Tomasz Kojm <zolw@konarski.edu.pl>\n"
"	  \n"
"    --help		    -h		show help\n"
"    --version		    -V		print version number and exit\n"
"    --verbose		    -v		be verbose\n"
"    --quiet				be quiet, output only error messages\n"
"    --stdout				write to stdout instead of stderr\n"
"					(this help is always written to stdout)\n"
"\n"
"    --daemon		    -d		run in daemon mode\n"
"    --checks=#n             -c #n       #n checks by day, 1 <= n <= 50\n"
"    --datadir=DIRECTORY                 download new database in DIRECTORY\n"
"    --log=FILE		    -l FILE	save download report in FILE\n"
"    --log-verbose			save additional informations\n"
"    --http-proxy=hostname[:port]	use proxy server hostname\n"
"    --daemon-notify[=/path/clamav.conf] send RELOAD command to clamd\n\n");

    exit(0);
}
