/*
 *  C implementation of the Aho-Corasic pattern matching algorithm. It's based
 *  on ScannerDaemon's Java version by Kurt Huwig and
 *  http://www-sr.informatik.uni-tuebingen.de/~buehler/AC/AC.html
 *  Thanks to Kurt Huwig for pointing me to this page.
 *
 *  Copyright (C) 2002 Tomasz Kojm <zolw@konarski.edu.pl>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#include "clamav.h"
#include "others.h"
#include "matcher.h"
#include "unrarlib.h"


int cli_addpatt(struct cl_node *root, struct patt *pattern)
{
	struct cl_node *pos, *next;
	int i;

    if(pattern->length < CL_MIN_LENGTH) {
	return CL_EPATSHORT;
    }

    pos = root;

    for(i = 0; i < CL_MIN_LENGTH; i++) {
	//+256
	next = pos->next[((unsigned char) pattern->pattern[i]) & 0xff];

	if(next == NULL) {
	    next = (struct cl_node *) cli_calloc(1, sizeof(struct cl_node));

	    if(!next)
		return CL_EMEM;
	    pos->next[((unsigned char) pattern->pattern[i]) & 0xff] = next;
	}

	pos = next;
    }

    pos->islast = 1;

    pattern->next = pos->list;
    pos->list = pattern;

    return 0;
}

struct nodelist *cli_bfsadd(struct nodelist *bfs, struct cl_node *n)
{
	struct nodelist *new;

    new = (struct nodelist *) cli_calloc(1, sizeof(struct nodelist));

    new->next = bfs;
    new->node = n;
    bfs = new;

    return bfs;
}

void cli_failtrans(struct cl_node *root)
{
	struct nodelist *bfs = NULL, *bfsnext = NULL, *pt, *prev;
	struct cl_node *n, *q, *qs, *r;
	int i;

    root->fail = NULL; /* h(root) = fail */

    for(i = 0; i < CL_NUM_CHILDS; i++) {
	/* for all childs of root */
	n = root->next[i];
	if(n) {
	    n->fail = root;
	    bfs = cli_bfsadd(bfs, n);
	}
    }

    while(bfs) {
	bfsnext = NULL;

	pt = bfs;

	while(pt) {
	    q = pt->node;

	    for(i = 0; i < CL_NUM_CHILDS; i++) {
		qs = q->next[i];

		if(qs) {
		    bfsnext = cli_bfsadd(bfsnext, qs);
		    r = q->fail;
		    while(r && (r->next[i] == NULL))
			r = r->fail;

		    if(r == NULL)
			qs->fail = root->next[i];
		    else
			qs->fail = r->next[i];
		}
	    }

	    pt = pt->next;
	}

	/* free the list */

	pt = bfs;
	while(pt) {
	    prev = pt;
	    pt = pt->next;
	    free(prev);
	}

	bfs = bfsnext;
    }
}

void cli_freepatt(struct patt *list)
{
	struct patt *handler, *prev;


    handler = list;

    while(handler) {
	free(handler->pattern);
	free(handler->virname);
	prev = handler;
	handler = handler->next;
	free(prev);
    }
}

void cl_freetrie(struct cl_node *root)
{
	int i, j;
	struct cl_node *next;


    for(i = 0; i < CL_NUM_CHILDS; i++) {
	next = root->next[i];
	if(next) {
	    for(j = 0; j < CL_NUM_CHILDS; j++)
		if(next->next[j])
		    cl_freetrie(next->next[j]);
	    cli_freepatt(next->list);
	    free(next);
	}
    }

    cli_freepatt(root->list);
    free(root);
}

void cl_buildtrie(struct cl_node *root)
{
    cli_failtrans(root);
    cli_fasttrie(root, root);
}

void cli_fasttrie(struct cl_node *n, struct cl_node *root)
{
	int i;
	struct cl_node *next, *q;


    for(i = 0; i < CL_NUM_CHILDS; i++) {
	next = n->next[i];
	if(next) {
	    n->trans[i] = next;
	    if(!n->islast)
		cli_fasttrie(next, root);
	} else {
	    q = n;
	    do {
		q = q->fail;
	    } while(q && (q->next[i] == NULL));
	    n->trans[i] = q ? q->next[i] : root;
	}
    }
}

int cl_scanbuff(const char *buffer, unsigned int length, char **virname, const struct cl_node *root)
{
	struct cl_node *current;
	struct patt *pt;
	int i, position, *partcnt;

    current = (struct cl_node *) root;

    partcnt = (int *) cli_calloc(root->partsigs + 1, sizeof(int));

    for(i = 0; i < length; i++)  {
	current = current->trans[(unsigned char) buffer[i] & 0xff];

	if(current->islast) {
	    position = i - CL_MIN_LENGTH + 1;

	    pt = current->list;
	    while(pt) {
		if(cli_findpos(buffer, position, length, pt)) {
		    if(pt->sigid) { /* it's a partial signature */
			//dbgmsg("Partial signature (id %d, part %d)\n", pt->sigid, pt->partno);
			if(partcnt[pt->sigid] + 1 == pt->partno) {
			    if(++partcnt[pt->sigid] == pt->parts) { /* last */
				*virname = pt->virname;
				free(partcnt);
				return CL_VIRUS;
			    }
			}
		    } else { /* old type signature */
			if(virname)
			    *virname = pt->virname;
			free(partcnt);
			return CL_VIRUS;
		    }
		}

		pt = pt->next;
	    }

	}
    }

    free(partcnt);
    return CL_CLEAN;
}

int cli_findpos(const char *buffer, int offset, int length, const struct patt *pattern)
{
	int bufferpos = offset + CL_MIN_LENGTH;
	int postfixend = offset + length;
	int i;


    for(i = CL_MIN_LENGTH; i < pattern->length; i++) {

	bufferpos %= length;

	if(bufferpos == postfixend)
	    return 0;

	if(pattern->pattern[i] != buffer[bufferpos])
	    return 0;

	bufferpos++;
    }

    return 1;
}
